import talib
import numpy as np
import pandas as pd
from collections import deque
from decimal import Decimal


class Indicator():
    pass


def SMA(input_data, timeperiod=100):
    result = []
    q = deque(maxlen=timeperiod)
    for price in input_data:
        q.append(price)
        if len(q) < timeperiod:
            result.append(np.NAN)
        else:
            sma = sum([item for item in q]) / Decimal(timeperiod)
            result.append(sma)
    return pd.Series(result, index=input_data.index)

'''
close = np.random.random(100)
output = talib.SMA(close, 10)
print(output, type(output))
result = SMA(close, 10)
print(result)
'''




from .currency import Currency, Currencies
from .ticker import Spot
from .exchange import Exchanges


class IndexVendor():
    def __init__(self, name: str):
        self.name = name


class IndexVendors():
    OKEx = IndexVendor('okex')


class Index():
    def __init__(self, weights: dict, base: Currency, vendor: IndexVendor):
        self.weights = weights
        self.base = base
        self.vendor = vendor


class Indices():
    BTCUSD = Index(
        {Spot(Currencies.BTC, Exchanges.Kraken): 1},
        Currencies.USD, IndexVendors.OKEx
    )
    ETHUSD = Index(
        {Spot(Currencies.ETH, Exchanges.Kraken): 1},
        Currencies.USD, IndexVendors.OKEx
    )
    EOSUSD = Index(
        {Spot(Currencies.EOS, Exchanges.Kraken): 1},
        Currencies.USD, IndexVendors.OKEx
    )
    XRPUSD = Index(
        {Spot(Currencies.XRP, Exchanges.Kraken): 1},
        Currencies.USD, IndexVendors.OKEx
    )



from abc import ABC, abstractmethod
from .account import Account
from .market_data import MarketData
from .currency import Currency


class Volume(ABC):
    @abstractmethod
    def get_volume(self, account: Account, market: MarketData):
        pass


class FixedVolume(Volume):
    def __init__(self, quantity: int, unit: Currency):
        self.quantity = quantity
        self.unit = unit

    def get_volume(self, account: Account, market: MarketData):
        return 0


class PercentVolume(Volume):
    def __init__(self, percent: float):
        self.percent = percent



from datetime import datetime
from decimal import Decimal
from typing import List

import pandas
import talib
import dateutil.parser as dp

from .chainable import Chainable
from .framework import Strategy
from .market_data import MarketData
from .ticker import Pair
from .common import Interval, Instructions
from .candle import Close
from .technical_analysis import MA
from .trade import Side
from .currency import Currency
from .indicator import SMA

# from .okex_trade import OkexTrade


class BacktestResult():
    pass


class BacktestConfig(Chainable):
    def __init__(
        self, pairs: List[Pair], interval: Interval,
        start: datetime, end: datetime,
        maker_fee: Decimal, taker_fee: Decimal
    ):
        self.pairs = pairs
        self.interval = interval
        self.start = start
        self.end = end
        self.maker_fee = maker_fee
        self.taker_fee = taker_fee

        # self.trade = OkexTrade("ETHUSD")


class Backtest():
    def __init__(self, config: BacktestConfig):
        self.config = config

    def set_config(self, config: BacktestConfig) -> None:
        self.config = config

    def get_config(self) -> BacktestConfig:
        return self.config

    def run(self, strategy: Strategy):
        net_asset_list = []
        net_asset = {
                        Currency('USD').name: 10,
                        Currency('USD').name + '_equity': 0.,
                        'equity': 0.
                    }
        for pair in self.config.pairs:
            net_asset[pair.quote_currency.name] = 10
            net_asset[pair.quote_currency.name + '_equity'] = 0.
        # net_asset_list.append([self.config.start, net_asset.copy()])

        ohlcs = {}
        for pair in self.config.pairs:
            ohlcs[pair] = MarketData().ohlc_history(
                pair, self.config.interval,
                self.config.start, self.config.end
            )
        #for pair, ohlc in ohlcs.items():
        #    print(ohlc, type(ohlc))
        #    r = talib.SMA(ohlc['close'], 100)
        #    print(r, type(r))
        df = pandas.concat({
            pair: pandas.concat([
                ohlc[['open', 'close']],
                SMA(ohlc['close'], 100).rename('ma_close_100')
                #talib.SMA(ohlc['close'], 100).rename('ma_close_100')
            ], axis=1)
            for pair, ohlc in ohlcs.items()
        }, axis=1)
        #print(df.iloc[:100])

        for index, row in df.iterrows():
            close_dict = {}
            for item in row.iteritems():
                close = Close(
                        pair=item[0][0],
                        interval=self.config.interval
                    )
                if item[0][1] == 'close':
                    close.set_value(item[1])
                elif item[0][1] == 'ma_close_100':
                    ma = MA(
                        close,
                        100
                    )
                    ma.set_value(item[1])
                close_dict[str(item[0][0])] = close

            try:
                strategy.on_tick()
            except Exception as e:
                print(e)
            for instruction in Instructions().get_all():
                pair = instruction.get_symbol().split('@')[0]
                side = instruction.get_side()
                status = instruction.get_status()
                print(index, close.get_key(), ma)
                print(instruction.symbol, instruction.side)
                print(pair)
                quote_name, base_name = pair.split('/')
                close = close_dict[instruction.get_symbol()].get_value()
                if instruction.side == Side.BUY:
                    net_asset[quote_name] += 1
                    net_asset[base_name] -= 1 * close
                elif instruction.side == Side.SELL:
                    net_asset[quote_name] -= 1
                    net_asset[base_name] += 1 * close
            net_asset['equity'] = 0
            net_asset['USD_equity'] = net_asset['USD']
            for pair in self.config.pairs:
                quote = pair.quote_currency.name
                close = close_dict[str(pair)].get_value()
                net_asset[quote + '_equity'] = close * net_asset[quote]

            for key in net_asset:
                if '_equity' in key:
                    net_asset['equity'] += net_asset[key]
            net_asset_list.append([index, net_asset.copy()])
            Instructions().clear()
        x, y = [], []
        for time, dic in net_asset_list:
            #x.append(dp.parse(time))
            x.append(time)
            # x.append(dp.parse(time).strftime('%d %H:%M'))
            y.append(dic['equity'])
            print(time, dic['equity'])
        '''
        import matplotlib.pyplot as plt
        plt.xlabel('datetime')
        plt.ylabel('net asset')
        plt.title('backtest')
        plt.gcf().autofmt_xdate()
        plt.xticks(rotation=90)
        plt.plot_date(x, y, color='b', linestyle='-', marker=',', label='okex')
        plt.legend(loc='best')
        plt.show()
        '''



from .ticker import *
from .trade import *
from .candle import *
from .technical_analysis import *
from .backtest import *
from .tradetest import *
from .currency import *
from .exchange import *
from .market_data import *
from .common import *
from .framework import *


from decimal import Decimal
from enum import Enum

from .ticker import Symbol
from .volume import Volume
from .common import Instructions, Instruction, Status


class Side(Enum):
    BUY = 'buy'
    SELL = 'sell'


class placeOrder(Instruction):
    def __init__(self, symbol, side, volume, status, price=None):
        Instructions().add(self)
        self.symbol = symbol.get_key()
        self.side = side.value
        self.volume = volume
        self.status = status.get()
        self.price = price

    def get_symbol(self):
        return self.symbol

    def get_side(self):
        return self.side

    def get_volume(self):
        return self.volume

    def get_status(self):
        return self.status
        

class placeLimitOrder(placeOrder):
    def __init__(
        self, symbol: Symbol, side: Side, volume: Volume, status: Status, price: Decimal
    ):
        super().__init__(symbol, side, volume, status, price)


class placeMarketOrder(placeOrder):
    def __init__(self, symbol: Symbol, side: Side, volume: Volume):
        super().__init__()
        self.symbol = symbol
        self.side = side
        self.volume = volume


class placeMakerOrder(placeOrder):
    def __init__(self, symbol: Symbol, side: Side, volume: Volume, status: Status):
        super().__init__(symbol, side, volume, status)


class placeTakerOrder(placeOrder):
    def __init__(self, symbol: Symbol, side: Side, volume: Volume, status: Status):
        super().__init__(symbol, side, volume, status)



class Currency():
    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return self.name


class Currencies():
    BTC = Currency('BTC')
    ETH = Currency('ETH')
    USDT = Currency('USDT')
    USD = Currency('USD')
    EOS = Currency('EOS')
    XRP = Currency('XRP')

    @staticmethod
    def get_currency(name: str):
        if name == 'BTC':
            return Currencies.BTC
        elif name == 'ETH':
            return Currencies.ETH
        elif name == 'EOS':
            return Currencies.EOS
        elif name == 'XRP':
            return Currencies.XRP
        elif name == 'USD':
            return Currencies.USD
        elif name == 'USDT':
            return Currencies.USDT
        else:
            return NotImplemented



import asyncio
import threading

from .okex import runClient
from .trade_tool import TradeTool
from .utf8logger import INFO


class OkexTrade(object):
    def __init__(self, pair):
        loop = asyncio.get_event_loop()
        inputs = asyncio.Queue(loop=loop)
        stop = loop.create_future()
        self.trade = TradeTool(loop, inputs)
        asyncio.ensure_future(runClient(loop, inputs, stop, self.trade, pair), loop=loop)
        thread = threading.Thread(target=loop.run_forever)
        thread.start()
        while self.trade.getLongQty() is None:
            pass
        INFO("trade is ready!")
        # try:
        #     while True:
        #         # Since there's no size limit, put_nowait is identical to put.
        #         message = input("> ")
        #         if message == '1': # ENTERLONG
        #             self.trade.do_maker_parallel("ENTER", "LONG", 5)
        #         elif message == '2': # ENTERSHORT
        #             self.trade.do_maker_parallel("ENTER", "SHORT", 5)
        #         elif message == '3': # EXITLONG
        #             self.trade.do_taker_parallel("EXIT", "LONG", 0)
        #         elif message == '4': # EXITSHORT
        #             self.trade.do_maker_parallel("EXIT", "SHORT", 0)
        #         elif message == '0': # quick empty
        #             self.trade.empty_quickly()
        #         else:
        #             continue
        # except (KeyboardInterrupt, EOFError):
        #     loop.call_soon_threadsafe(stop.set_result, None)

    def enter_long(self, typ, vol):
        if typ == "taker":
            self.trade.do_taker_parallel("ENTER", "LONG", vol)
        if typ == "maker":
            self.trade.do_maker_parallel("ENTER", "LONG", vol)

    def enter_short(self, typ, vol):
        if typ == "taker":
            self.trade.do_taker_parallel("ENTER", "SHORT", vol)
        if typ == "maker":
            self.trade.do_maker_parallel("ENTER", "SHORT", vol)

    def exit_long(self, typ, vol):
        if typ == "taker":
            self.trade.do_taker_parallel("EXIT", "LONG", vol)
        if typ == "maker":
            self.trade.do_maker_parallel("EXIT", "LONG", vol)

    def exit_short(self, typ, vol):
        if typ == "taker":
            self.trade.do_taker_parallel("EXIT", "SHORT", vol)
        if typ == "maker":
            self.trade.do_maker_parallel("EXIT", "SHORT", vol)

    def empty_all(self):
        self.trade.empty_quickly()

if __name__ == "__main__":
    a = OkexTrade("ETHUSD")
    # a.enter_long("taker", 5)



from datetime import datetime

import requests
import pandas
from mathlab_pg_api.futures_api import FuturesApi
from mathlab_pg_api.spot_api import SpotApi

from .exchange import Exchanges
from .common import Interval, MarketType
from .ticker import Pair, Symbol
from .singleton import Singleton

#POSTGREST_URL = 'http://localhost:3000/test2'


class MarketData(Singleton):
    def __init__(self):
        self.futures_api = FuturesApi('localhost', '5678')
        self.spot_api = SpotApi('localhost', '5678')

    # Old version
    #def ohlc_history(
    #    self, pair: Pair, interval: Interval,
    #    start: datetime, end: datetime
    #) -> pandas.DataFrame:
    #    if pair.get_exchange() != Exchanges.Binance:
    #        return NotImplemented
    #    if not isinstance(pair, Pair):
    #        return NotImplemented
    #    if interval != Interval.M1:
    #        return NotImplemented
    #    params = {
    #        'and': '(time.gte.%s,time.lte.%s)' % (
    #            start.isoformat(), end.isoformat()),
    #        'pair': 'eq.%s%s' % (pair.quote_currency, pair.base_currency),
    #        'exchange_id': 'eq.6',
    #        'period': 'eq.1',
    #        'order': 'time'
    #    }
    #    r = requests.get(POSTGREST_URL, params=params)
    #    if r.status_code == 200:
    #        df = pandas.DataFrame(r.json())
    #        df.set_index('time', inplace=True)
    #        return df
    #    return NotImplemented

    def ohlc_history(
        self, pair: Pair, interval: Interval,
        start: datetime, end: datetime, market_type = MarketType.Futures
    ) -> pandas.DataFrame:
        if market_type == MarketType.Futures:
            api = self.futures_api
        else:
            api = self.spot_api
        return api.candles(exchange_name=pair.get_exchange(), pair=pair.get_pair(), period=interval.value,
                                        start_date=start, end_date=end)



from .common import Parameters
from abc import ABC, abstractmethod


class Strategy(ABC):
    def __init__(self, **kargs):
        params = Parameters()
        for key, value in kargs.items():
            params.set(key, value)

    @abstractmethod
    def on_tick(self):
        return NotImplemented



# -*- coding: utf-8 -*-

import sys, os
p = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
if p not in sys.path:
    sys.path.insert(0, p)

import sys, logging

PY3 = sys.version_info[0] == 3

def equalUtf8(coding):
    return coding is None or coding.lower() in ('utf8', 'utf-8', 'utf_8')

class CodingWrappedWriter(object):
    def __init__(self, coding, writer):
        self.flush = getattr(writer, 'flush', lambda : None)
        
        wcoding = getattr(writer, 'encoding', None)
        wcoding = 'gb18030' if (wcoding in ('gbk', 'cp936')) else wcoding

        if not equalUtf8(wcoding):
            self._write = lambda s: writer.write(
                s.decode(coding).encode(wcoding, 'ignore')
            )
        else:
            self._write = writer.write
    
    def write(self, s):
        self._write(s)
        self.flush()

if not PY3:
    # utf8Stdout.write("中文") <==> 
    # sys.stdout.write("中文".decode('utf8').encode(sys.stdout.encoding))
    utf8Stdout = CodingWrappedWriter('utf8', sys.stdout)
else:
    # reference http://blog.csdn.net/jim7424994/article/details/22675759
    import io
    if hasattr(sys.stdout, 'buffer') and (not equalUtf8(sys.stdout.encoding)):
        if sys.stdout.encoding in ('gbk', 'cp936'):
            coding = 'gb18030'
        else:
            coding = 'utf-8'
        utf8Stdout = io.TextIOWrapper(sys.stdout.buffer, encoding=coding)
    else:
        utf8Stdout = sys.stdout

#class _utf8Stdout(object):
#
#    @classmethod
#    def write(cls, s):
#        utf8Stdout.write(s)
#        utf8Stdout.flush()
#
#    @classmethod
#    def flush(cls):
#        utf8Stdout.flush()

def Utf8Logger(name):
    logger = logging.getLogger(name)
    if not logger.handlers:
        logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler(utf8Stdout)
        fmt = '[%(asctime)s] [%(levelname)s] %(message)s'
        datefmt = '%Y-%m-%d %H:%M:%S'
        ch.setFormatter(logging.Formatter(fmt, datefmt))
        logger.addHandler(ch)
    return logger

logging.getLogger("").setLevel(logging.DEBUG)

utf8Logger = Utf8Logger('Utf8Logger')

def SetLogLevel(level):
    utf8Logger.setLevel(getattr(logging, level.upper()))

def DisableLog():
    utf8Logger.disabled = True

def EnableLog():
    utf8Logger.disabled = False

CRITICAL=utf8Logger.critical
ERROR=utf8Logger.error
WARN=utf8Logger.warn
INFO=utf8Logger.info
DEBUG=utf8Logger.debug


if PY3:
    RAWINPUT = input
else:
    def RAWINPUT(msg):
        utf8Stdout.write(msg)
        s = raw_input('').rstrip()
        if not equalUtf8(sys.stdin.encoding):
            s = s.decode(sys.stdin.encoding).encode('utf8')
        return s        

def PRINT(s, end='\n'):
    utf8Stdout.write(s+end)
    utf8Stdout.flush()

def test():
    s = RAWINPUT("请输入一串中文：")
    PRINT(s)
    INFO(s)
    CRITICAL(s)

if __name__ == '__main__':
    test()
    


from datetime import datetime

from .exchange import Exchange
from .index import Index
from .ticker import Symbol


class IndexFutures(Symbol):
    def __init__(
        self, index: Index, expire_date: datetime, exchange: Exchange
    ):
        self.index = index
        self.expire_date = expire_date
        self.exchange = exchange



from .ticker import Pair
from .common import Value, Parameter, Interval


class Open():
    pass


class High():
    pass


class Low():
    pass


class Close(Value):
    def __init__(
        self,
        pair: Pair = Parameter('pair'),
        interval: Interval = Parameter('interval'),
        shift: int = 0
    ):
        super().__init__(
            key='close_%s_%s_%s' % (pair.get_key(), interval, shift))



import itertools

try:
    import builtins
except ImportError:
    import __builtin__ as builtins


class Chainable(object):
    def __init__(self, data, method=None):
        self.data = data
        self.method = method

    def __getattr__(self, name):
        try:
            method = getattr(self.data, name)
        except AttributeError:
            try:
                method = getattr(builtins, name)
            except AttributeError:
                method = getattr(itertools, name)

        return Chainable(self.data, method)

    def __call__(self, *args, **kwargs):
        try:
            return Chainable(list(self.method(self.data, *args, **kwargs)))
        except TypeError:
            return Chainable(list(self.method(args[0], self.data, **kwargs)))



import time
from .utf8logger import INFO, ERROR


class TradeTool(object):
    def __init__(self, loop, inputs):
        self.side_dic = {"ENTERLONG": 1, "ENTERSHORT": 2, "EXITLONG": 3, "EXITSHORT": 4}
        self.loop = loop
        self.inputs = inputs
        self.stall_volume = 1
        self.long_qty = None
        self.short_qty = None
        self.long_avail_qty = None
        self.short_avail_qty = None
        self.order_status_map = {}

    def do_maker_parallel(self, action, side, current_volume):
        action_side = action + side
        name = "maker"
        if action_side == "ENTERLONG":
            while current_volume > self.long_avail_qty:
                vol_list = set()
                tmp_volume = current_volume
                while tmp_volume > self.long_avail_qty:
                    real_vol = min(self.stall_volume, tmp_volume - self.long_avail_qty)
                    vol_list.add(real_vol)
                    tmp_volume -= self.stall_volume
                self.epoch_parallel(name, action_side, vol_list)

        if action_side == "ENTERSHORT":
            while current_volume > self.short_avail_qty:
                vol_list = set()
                tmp_volume = current_volume
                while tmp_volume > self.short_avail_qty:
                    real_vol = min(self.stall_volume, tmp_volume - self.short_avail_qty)
                    vol_list.add(real_vol)
                    tmp_volume -= self.stall_volume
                self.epoch_parallel(name, action_side, vol_list)

        if action_side == "EXITLONG":
            while self.long_qty > current_volume:
                vol_list = set()
                tmp_volume = self.long_qty
                while tmp_volume > current_volume:
                    real_vol = min(self.stall_volume, tmp_volume - current_volume)
                    vol_list.add(real_vol)
                    tmp_volume -= self.stall_volume
                self.epoch_parallel(name, action_side, vol_list)

        if action_side == "EXITSHORT":
            while self.short_qty > current_volume:
                vol_list = set()
                tmp_volume = self.short_qty
                while tmp_volume > current_volume:
                    real_vol = min(self.stall_volume, tmp_volume - current_volume)
                    vol_list.add(real_vol)
                    tmp_volume -= self.stall_volume
                self.epoch_parallel(name, action_side, vol_list)

    def do_taker_parallel(self, action, side, current_volume):
        action_side = action + side
        name = "taker"
        if action_side == "ENTERLONG":
            while current_volume > self.long_avail_qty:
                vol_list = set()
                tmp_volume = current_volume
                while tmp_volume > self.long_avail_qty:
                    real_vol = min(self.stall_volume, tmp_volume - self.long_avail_qty)
                    vol_list.add(real_vol)
                    tmp_volume -= self.stall_volume
                self.epoch_parallel(name, action_side, vol_list)

        if action_side == "ENTERSHORT":
            while current_volume > self.short_avail_qty:
                vol_list = set()
                tmp_volume = current_volume
                while tmp_volume > self.short_avail_qty:
                    real_vol = min(self.stall_volume, tmp_volume - self.short_avail_qty)
                    vol_list.add(real_vol)
                    tmp_volume -= self.stall_volume
                self.epoch_parallel(name, action_side, vol_list)

        if action_side == "EXITLONG":
            while self.long_qty > current_volume:
                vol_list = set()
                tmp_volume = self.long_qty
                while tmp_volume > current_volume:
                    real_vol = min(self.stall_volume, tmp_volume - current_volume)
                    vol_list.add(real_vol)
                    tmp_volume -= self.stall_volume
                self.epoch_parallel(name, action_side, vol_list)

        if action_side == "EXITSHORT":
            while self.short_qty > current_volume:
                vol_list = set()
                tmp_volume = self.short_qty
                while tmp_volume > current_volume:
                    real_vol = min(self.stall_volume, tmp_volume - current_volume)
                    vol_list.add(real_vol)
                    tmp_volume -= self.stall_volume
                self.epoch_parallel(name, action_side, vol_list)

    def epoch_parallel(self, name, side, vol_list):
        order_nums = len(vol_list)
        INFO(name + " parallel: " + side + ", order counts: {}".format(order_nums))
        for real_vol in vol_list:
            self.loop.call_soon_threadsafe(
                self.inputs.put_nowait,
                (self.side_dic[side], name, real_vol)
            )
        INFO(side + " parallel waiting for response...")
        max_waiting_secs = 3
        t = 0
        while len(self.order_status_map) < order_nums:
            INFO(side + " waiting for parallel order status changed...")
            time.sleep(1)
            t += 1
            if t >= max_waiting_secs:
                ERROR(name + " parallel: out of waiting time {} secs".format(t))
                break
            pass
        for oid in self.order_status_map:
            if self.order_status_map[oid] in (0, 1):
                self.loop.call_soon_threadsafe(
                    self.inputs.put_nowait,
                    (oid, "cancel", None)
                )
        self.order_status_map = {}

    def empty_quickly(self):
        if self.long_qty > 0:
            self.do_taker_parallel("EXIT", "LONG", 0)
        if self.short_qty > 0:
            self.do_taker_parallel("EXIT", "SHORT", 0)

    def getStallVolume(self):
        return self.stall_volume

    def setLongQty(self, qty):
        self.long_qty = int(qty)

    def getLongQty(self):
        return self.long_qty

    def setShortQty(self, qty):
        self.short_qty = int(qty)

    def setLongAvailQty(self, qty):
        self.long_avail_qty = int(qty)

    def setShortAvailQty(self, qty):
        self.short_avail_qty = int(qty)

    def setOrderStatus(self, os, oid):
        self.order_status_map[oid] = os



class _Singleton(type):
    """ A metaclass that creates a Singleton base class when called. """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(
                _Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Singleton(_Singleton('SingletonMeta', (object,), {})):
    pass



#!/usr/bin/env python

# WS client example

import asyncio
import base64
import hmac
import json
import os
import sys
import threading
import time
import zlib
import aiohttp
import requests
import websockets
import datetime

from .trade_tool import TradeTool
from .utf8logger import INFO, WARN


def print_during_input(string):
    sys.stdout.write(
        # Save cursor position
        "\N{ESC}7"
        # Add a new line
        "\N{LINE FEED}"
        # Move cursor up
        "\N{ESC}[A"
        # Insert blank line, scroll last line down
        "\N{ESC}[L"
        # Print string in the inserted blank line
        f"{string}\N{LINE FEED}"
        # Restore cursor position
        "\N{ESC}8"
        # Move cursor down
        "\N{ESC}[B"
    )
    sys.stdout.flush()

class OKEx():
    def __init__(self, loop, strategy, pair):
        self.loop = loop
        self.instrument_id = None
        # self.key = '1957674e-8d18-4c24-bd35-8c80fbaebccf'
        # self.secret = 'C1F9D221E1AC8D990CD08C8B67CA5746'
        self.key = 'e3df3679-409b-42d7-b7f7-22f834a6c9cb'
        self.secret = 'AB300727C16C38BAFEF4B35DA0DA92CC'
        self.passphrase = 'Lol2018'
        self.host = 'https://www.okex.com'
        self.CONTENT_TYPE = 'Content-TYPE'
        self.OK_ACCESS_KEY = 'OK-ACCESS-KEY'
        self.OK_ACCESS_SIGN = 'OK-ACCESS-SIGN'
        self.OK_ACCESS_TIMESTAMP = 'OK-ACCESS-TIMESTAMP'
        self.OK_ACCESS_PASSPHRASE = 'OK-ACCESS-PASSPHRASE'
        self.APPLICATION_JSON = 'application/json'
        self.bid = None
        self.ask = None
        self.strategy = strategy
        self.pair = pair
        self.OKEX_PAIR_MAP = {
            'ETHUSD': {
                'underlying_index': 'ETH',
                'quote_currency': 'USD',
            },
            'BTCUSD': {
                'underlying_index': 'BTC',
                'quote_currency': 'USD',
            },
            'EOSUSD': {
                'underlying_index': 'EOS',
                'quote_currency': 'USD',
            },
            'XRPUSD': {
                'underlying_index': 'XRP',
                'quote_currency': 'USD',
            },
            'LTCUSD': {
                'underlying_index': 'LTC',
                'quote_currency': 'USD',
            },
            'ETCUSD': {
                'underlying_index': 'ETC',
                'quote_currency': 'USD',
            },
        }

    def inflate(self, data):
        decompress = zlib.decompressobj(
                -zlib.MAX_WBITS  # see above
        )
        inflated = decompress.decompress(data)
        inflated += decompress.flush()
        return inflated.decode('utf8')

    def signature(self, timestamp, method, request_path, body):
        if str(body) == '{}' or str(body) == 'None':
            body = ''
        message = str(timestamp) + str.upper(method) + request_path + str(body)
        mac = hmac.new(bytes(self.secret, encoding='utf8'), bytes(message, encoding='utf-8'), digestmod='sha256')
        d = mac.digest()
        return base64.b64encode(d).decode('utf8')

    def get_header(self, sign, timestamp):
        header = dict()
        header[self.CONTENT_TYPE] = self.APPLICATION_JSON
        header[self.OK_ACCESS_KEY] = self.key
        header[self.OK_ACCESS_SIGN] = sign
        header[self.OK_ACCESS_TIMESTAMP] = str(timestamp)
        header[self.OK_ACCESS_PASSPHRASE] = self.passphrase
        return header

    async def __aenter__(self):
        self._conn = websockets.connect('wss://real.okex.com:10442/ws/v3')
        self.websocket = await self._conn.__aenter__()
        self.session = aiohttp.ClientSession()
        return self

    async def __aexit__(self, *args, **kwargs):
        await self._conn.__aexit__(*args, **kwargs)

    async def send(self, message):
        print_during_input("< " + message)
        await self.websocket.send(message)

    async def receive(self):
        received = await self.websocket.recv()
        return self.inflate(received)

    async def login(self):
        timestamp = '%.3f' % time.time()
        login = {'op': 'login', 'args': [self.key, self.passphrase, timestamp, self.signature(timestamp, 'GET', '/users/self/verify', '')]}
        await self.send(json.dumps(login))

    async def subscribe(self):
        self.instrument_id = self.get_latest_future_instrument_id(self.pair)
        INFO("subscribe instrument id: {}".format(self.instrument_id))
        subscribe = {
            'op': 'subscribe',
            'args': [
                'futures/ticker:' + self.instrument_id,
                'futures/position:' + self.instrument_id,
                'futures/account:' + self.OKEX_PAIR_MAP[self.pair]['underlying_index'],
                'futures/order:' + self.instrument_id,
                'spot/account:' + self.OKEX_PAIR_MAP[self.pair]['underlying_index']
            ]
        }
        await self.send(json.dumps(subscribe))

    async def request(self, path, req={}, method='get'):
        timestamp = '%.3f' % time.time()
        data_str = ""
        if req:
            url = '?'
            for key, value in req.items():
                url = url + str(key) + '=' + str(value) + '&'
            path += url[0: -1]
            data_str = json.dumps(req)
        urlpath = os.path.join(self.host, path)
        if method == "get":
            header = self.get_header(self.signature(timestamp, 'GET', '/' + path, data_str), timestamp)
            r = await self.session.get(urlpath, headers=header)
        elif method == 'post':
            header = self.get_header(self.signature(timestamp, 'POST', '/' + path, data_str), timestamp)
            r = await self.session.post(urlpath, data=data_str, headers=header)
        return await r.json()

    def parseTicker(self, data):
        if 'best_bid' in data and self.bid != data['best_bid']:
            self.bid = data['best_bid']
        if 'best_ask' in data and self.ask != data['best_ask']:
            self.ask = data['best_ask']

    def get_latest_future_instrument_id(self, pair):
        while 1:
            try:
                response = requests.get('https://www.okex.com/api/futures/v3/instruments', timeout=3)
                instruments = response.json()
                break
            except Exception as e:
                WARN('OKEX instruments request timeout' + str(e))
                continue
        future_delivery_date = ''
        future_instrument_id = ''
        for instrument in instruments:
            if instrument['underlying_index'] == self.OKEX_PAIR_MAP[pair]['underlying_index'] and \
                    instrument['quote_currency'] == self.OKEX_PAIR_MAP[pair]['quote_currency']:
                listing = instrument['listing'].split('-')
                delivery = instrument['delivery'].split('-')
                listing_date = datetime.datetime(int(listing[0]), int(listing[1]), int(listing[2]))
                delivery_date = datetime.datetime(int(delivery[0]), int(delivery[1]), int(delivery[2]))
                if (delivery_date - listing_date).days < 90:
                    continue
                if (not future_delivery_date) or int(time.mktime(delivery_date.timetuple())) > int(time.mktime(time.strptime(future_delivery_date, "%Y-%m-%d"))):
                    future_instrument_id = instrument['instrument_id']
        return future_instrument_id

    async def getWallet(self):
        r = await self.request(path = 'api/spot/v3/accounts/' + self.OKEX_PAIR_MAP[self.pair]['underlying_index'] + '/ledger')
        print_during_input(r)

    async def getPosition(self):
        self.instrument_id = self.get_latest_future_instrument_id(self.pair)
        r = await self.request(path = 'api/futures/v3/{0}/position'.format(self.instrument_id))
        print_during_input(r)
        if r['holding']:
            self.parsePosition(r['holding'][0])

    def parsePosition(self, data):
        if 'long_avail_qty' in data:
            self.strategy.setLongAvailQty(data['long_avail_qty'])
        if 'short_avail_qty' in data:
            self.strategy.setShortAvailQty(data['short_avail_qty'])
        if 'long_qty' in data:
            self.strategy.setLongQty(data['long_qty'])
        if 'short_qty' in data:
            self.strategy.setShortQty(data['short_qty'])

    def parseOrder(self, data):
        order_status = int(data["status"])
        order_id = data["order_id"]
        prefix = "type {0}, order_id {1},".format(data["type"], order_id)
        self.strategy.setOrderStatus(order_status, order_id)
        if order_status == 0:
            INFO(prefix + " order committed")
        elif order_status == 2:
            INFO(prefix + " order completed")
        else:
            INFO(prefix + " do nothing, status {0}".format(order_status))

    async def parse(self, message):
        body = json.loads(message)
        if 'event' in body:
            if 'login' == body['event'] and body['success']:
                asyncio.ensure_future(self.subscribe(), loop = self.loop)
            print_during_input("> " + message)
        elif 'table' in body:
            if 'futures/ticker' == body['table']:
                for item in body['data']:
                    self.parseTicker(item)
            if 'futures/position' == body['table']:
                self.parsePosition(body['data'][0])
                print_during_input("> " + message)
            if 'futures/order' == body['table']:
                for item in body['data']:
                    self.parseOrder(item)
                print_during_input("> " + message)
        else:
            print_during_input(f'Unknown message {message}')

    async def placeMakerOrder(self, side, vol):
        if side in [2, 3]:
            price = self.ask
        elif side in [1, 4]:
            price = self.bid
        r = await self.request('api/futures/v3/order', req={'instrument_id': self.instrument_id, 'type': side, 'price': price, 'size': vol, 'leverage': 20}, method='post')
        print_during_input(r)
        return r

    async def placeTakerOrder(self, side, vol):
        if side in [1, 4]:
            price = self.ask
        elif side in [2, 3]:
            price = self.bid
        r = await self.request('api/futures/v3/order', req={'instrument_id': self.instrument_id, 'type': side, 'price': price, 'size': vol, 'leverage': 20}, method='post')
        print_during_input(r)
        return r

    async def cancelOrder(self, oid):
        r = await self.request(
            'api/futures/v3/cancel_order/{0}/{1}'.format(self.instrument_id, oid),
            req={},
            method='post'
        )
        print_during_input(r)
        return r

async def runClient(loop, inputs, stop, strategy, pair="ETHUSD"):
    async with OKEx(loop, strategy, pair) as ex:
        asyncio.ensure_future(ex.login())
        asyncio.ensure_future(ex.getPosition())
        # asyncio.ensure_future(ex.getWallet())

        while True:
            incoming = asyncio.ensure_future(ex.receive(), loop=loop)
            outgoing = asyncio.ensure_future(inputs.get(), loop=loop)
            done, pending = await asyncio.wait(
                [incoming, outgoing, stop], return_when=asyncio.FIRST_COMPLETED
            )

            if incoming in pending:
                incoming.cancel()
            if outgoing in pending:
                outgoing.cancel()

            if incoming in done:
                try:
                    message = incoming.result()
                    asyncio.ensure_future(ex.parse(message), loop=loop)
                except Exception as e:
                    print_during_input(e)
                    break

            if outgoing in done:
                para_1, means, para_2 = outgoing.result()
                if means == "taker":
                    asyncio.ensure_future(ex.placeTakerOrder(para_1, para_2), loop=loop)
                elif means == "maker":
                    asyncio.ensure_future(ex.placeMakerOrder(para_1, para_2), loop=loop)
                elif means == "cancel":
                    asyncio.ensure_future(ex.cancelOrder(para_1), loop=loop)

            if stop in done:
                break
        print_during_input('Client Shutdown!')


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    inputs = asyncio.Queue(loop=loop)
    stop = loop.create_future()
    strategy = TradeTool(loop, inputs)
    asyncio.ensure_future(runClient(loop, inputs, stop, strategy), loop=loop)
    thread = threading.Thread(target=loop.run_forever)
    thread.start()
    try:
        while True:
            # Since there's no size limit, put_nowait is identical to put.
            message = input("> ")
            if message == 'open long':
                side = 1
            elif message == 'open short':
                side = 2
            elif message == 'close long':
                side = 3
            elif message == 'close short':
                side = 4
            else:
                continue
            loop.call_soon_threadsafe(inputs.put_nowait, (side, "taker", 1))
    except (KeyboardInterrupt, EOFError):  # ^C, ^D
        loop.call_soon_threadsafe(stop.set_result, None)



class Exchange():
    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return self.name


class Exchanges():
    Binance = Exchange('binance')
    Kraken = Exchange('kraken')
    OKEX = Exchange('okex')



import setuptools

with open("README.md", "r") as fh:
        long_description = fh.read()

        setuptools.setup(
            name="mathlab_strategy-andrete",
            version="0.0.2",
            author="andrete",
            author_email="13120309879@163.com",
            description="mathlab strategy framework",
            long_description=long_description,
            long_description_content_type="text/markdown",
            url="https://github.com/mathlabchina/Pythagoras",
            packages=setuptools.find_packages(),
            classifiers=[
                "Programming Language :: Python :: 3",
                "License :: OSI Approved :: MIT License",
                "Operating System :: OS Independent",
            ],
            install_requires=[
                'pandas',
                'TA-Lib',
                'requests'
            ]

        )



from datetime import datetime
from decimal import Decimal
from typing import List

import pandas
import talib

from .chainable import Chainable
from .framework import Strategy
from .market_data import MarketData
from .ticker import Pair
from .common import Interval, Instructions
from .candle import Close
from .technical_analysis import MA
from .trade import *

from .okex_trade import OkexTrade

class TradetestResult():
    pass


class TradetestConfig(Chainable):
    def __init__(
        self, pairs: List[Pair], interval: Interval,
        start: datetime, end: datetime,
        maker_fee: Decimal, taker_fee: Decimal
    ):
        self.pairs = pairs
        self.interval = interval
        self.start = start
        self.end = end
        self.maker_fee = maker_fee
        self.taker_fee = taker_fee


class Tradetest():
    def __init__(self, config: TradetestConfig):
        self.config = config
        self.trade = OkexTrade("ETHUSD")

    def set_config(self, config: TradetestConfig) -> None:
        self.config = config

    def get_config(self) -> TradetestConfig:
        return self.config

    def run(self, strategy: Strategy):
        ohlcs = {}
        for pair in self.config.pairs:
            ohlcs[pair] = MarketData().ohlc_history(
                pair, self.config.interval,
                self.config.start, self.config.end
            )
        df = pandas.concat({
            pair: pandas.concat([
                ohlc[['open', 'close']],
                talib.SMA(ohlc['close'], 100).rename('ma_close_100')
            ], axis=1)
            for pair, ohlc in ohlcs.items()
        }, axis=1)

        trade_counts = 0
        for index, row in df.iterrows():
            for item in row.iteritems():
                close = Close(
                        pair=item[0][0],
                        interval=self.config.interval
                )
                if item[0][1] == 'close':
                    close.set_value(item[1])
                elif item[0][1] == 'ma_close_100':
                    ma = MA(
                        close,
                        100
                    )
                    ma.set_value(item[1])
            try:
                strategy.on_tick()
            except Exception as e:
                print(e)
            for instruction in Instructions().get_all():
                pair = instruction.get_symbol().split('@')[0].replace('/', '')
                side = instruction.get_side()
                status = instruction.get_status()
                print(index, type(instruction), pair, side, status)

                way = None
                if isinstance(instruction, placeTakerOrder):
                    way = "taker"
                elif isinstance(instruction, placeMakerOrder):
                    way = "taker"

                if way is None:
                    continue

                trade_counts += 1
                if side == "buy":
                    if status == "long":
                        print(index, "open long...")
                        self.trade.enter_long(way, 1)
                    elif status == "short":
                        print(index, "open short...")
                        self.trade.enter_short(way, 1)
                elif side == "sell":
                    if status == "long":
                        print(index, "exit long...")
                        self.trade.exit_long(way, 0)
                    elif status == "short":
                        print(index, "exit short...")
                        self.trade.exit_short(way, 0)

            Instructions().clear()
            if trade_counts > 3:
                break



from enum import Enum
from typing import Any, Type, List

from .singleton import Singleton


class Instruction():
    pass


class Instructions(Singleton):
    def __init__(self):
        self.instructions = []

    def add(self, instruction: Instruction):
        self.instructions.append(instruction)

    def get_all(self) -> list:
        return self.instructions

    def clear(self) -> None:
        self.instructions = []


class Values(Singleton):
    def __init__(self):
        self.values = {}

    def set(self, key, new):
        self.values[key] = new

    def get(self, key):
        return self.values[key]

    def contains(self, key):
        return key in self.values.keys()


class Value():
    def __init__(self, key: str):
        self.key = key
        if not Values().contains(key):
            Values().set(key, None)

    def set_value(self, value: Any):
        Values().set(self.key, value)

    def get_value(self):
        return Values().get(self.key)

    def get_key(self):
        return self.key

    def __add__(self, right: Any):
        return self.get_value() + right

    def __radd__(self, left: Any):
        return left + self.get_value()

    def __sub__(self, right: Any):
        return self.get_value() - right

    def __rsub__(self, left: Any):
        return left - self.get_value()

    def __lt__(self, right: Any):
        return self.get_value() < right

    def __gt__(self, right: Any):
        return self.get_value() > right

    def __eq__(self, right: Any):
        return self.get_value() == right

    def __str__(self):
        return self.get_value().__str__()


class Parameters(Singleton):
    def __init__(self):
        self.parameters = {}

    def set(self, name: str, value: Any):
        print(name, value)
        self.parameters[name] = value

    def get(self, name: str):
        return self.parameters[name]

    def contains(self, name: str):
        return name in self.parameters.keys()


class ParameterFunc():
    def __call__(self, name: str, cls: Type = str, default: Any = None):
        if Parameters().contains(name):
            return Parameters().get(name)
        else:
            return default


Parameter = ParameterFunc()


class StatusEnv(Singleton):
    def __init__(self):
        self.envs = {}

    def set(self, name, value):
        self.envs[name] = value

    def get(self, name):
        return self.envs.get(name)

    def contains(self, name):
        return name in self.envs.keys()


class Status():
    def __init__(self, name: str, enum: List[str], init: str):
        self.name = name
        if not StatusEnv().contains(self.name):
            StatusEnv().set(self.name, init)

    def __is__(self, s: str) -> bool:
        return self.get() == s

    def __ne__(self, s) -> bool:
        if isinstance(s, str):
            return self.get() != s
        return NotImplemented

    def __eq__(self, s) -> bool:
        if isinstance(s, str):
            return self.get() == s
        return NotImplemented

    def set(self, s: str):
        StatusEnv().set(self.name, s)

    def get(self):
        return StatusEnv().get(self.name)


class Interval(Enum):
    M1 = 1 #'m1'
    M5 = 5 #'m5'
    M10 = 10 #'m10'
    M15 = 15 #'m15'
    M30 = 30 #'m30'
    Hour = 60 #'h1'
    Day = 1440 #'day'
    Week = 10080 #'week'

class MarketType(Enum):
    Futures = 1
    Spot = 2



from .currency import Currency
from .exchange import Exchange

from abc import ABC, abstractmethod


class Symbol(ABC):
    @abstractmethod
    def get_exchange(self):
        return NotImplemented

    @abstractmethod
    def get_key(self):
        return NotImplemented


class Spot(Symbol):
    def __init__(self, currency: Currency, exchange: Exchange):
        self.currency = currency
        self.exchange = exchange

    def get_exchange(self):
        return self.exchange

    def get_currency(self):
        return self.currency


class Pair(Symbol):
    def __init__(self, quote: Currency, base: Currency, exchange: Exchange):
        self.quote_currency = quote
        self.base_currency = base
        self.exchange = exchange

    def get_exchange(self):
        return self.exchange

    def get_key(self):
        return str(self)

    def get_pair(self):
        return '%s/%s' % (
            self.quote_currency,
            self.base_currency
        )

    def __str__(self):
        return '%s/%s@%s' % (
            self.quote_currency,
            self.base_currency,
            self.exchange
        )

    def __lt__(self, right):
        if isinstance(right, Pair):
            return str(self) < str(right)
        return NotImplemented
